# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import attr

@attr.s(cmp=False)
class OpType(object):
    name = attr.ib()
    args = attr.ib()
    projs = attr.ib()
    opcls = attr.ib()
    stararg = attr.ib()
    is_phi = attr.ib()
    source_type = attr.ib(default=None)

@attr.s(cmp=False)
class ProjType(object):
    name = attr.ib()
    mode = attr.ib()

def proj_definer(mode):
    def produce(name):
        return ProjType(name=name, mode=mode)
    return produce

M = ProjType(name="M", mode="M")
V = ProjType(name="V", mode="V")
C = ProjType(name="C", mode="C")
T = ProjType(name="T", mode="C")
F = ProjType(name="F", mode="C")


def ops(ops):
    optypes = {}
    more_than_one = []
    for op in ops:
        if op.name in optypes:
            more_than_one.append(op.name)
        optypes[op.name] = op
    if more_than_one:
        raise Exception(', '.join(sorted(more_than_one)))
    return optypes

def op_definer(opcls, standard_proj):
    def op(name, argnames=(), proj=None, star=None, is_phi=False,
           source_type=None):
        if proj is None:
            proj = standard_proj
        return OpType(name=name, args=list(argnames), projs=proj,
                      opcls=opcls, stararg=star, is_phi=is_phi,
                      source_type=source_type)
    return op

def op_language(sublanguages):
    result = {}
    for language in sublanguages:
        result.update(language)
    return result


op = op_definer("entangled", [])

initial_types = ops([
    op('UNBOUND', [], [V, M]),
    op('entangled.arg', ['number'], [V]),
    op('entangled.initial_memory', [], [M]),
    op('entangled.start', [], [V, M, C]),
])

op = op_definer("entangled.phi", [])

phi_types = ops([
    op('entangled.phi', [], [V], star='sources', is_phi=True),
    op('entangled.phi.stack', ['number'], [V], star='sources', is_phi=True),
    op('entangled.phi.local', ['number'], [V, M], star='sources', is_phi=True),
    op('entangled.phi.mem', [], [M], star='sources', is_phi=True),
    op('entangled.phi.exc_info', ['number'], [
        ProjType(name='EX_TY', mode='V'),
        ProjType(name='EX_V', mode='V'),
        ProjType(name='EX_TB', mode='V'),
        ProjType(name='EX_M', mode='M'),
    ], star='sources', is_phi=True),
    op('entangled.phi.retval', [], star='sources', is_phi=True),
])

conditions = ops([
    op('entangled.true', [], [C]),
])
