# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.op_types import (ProjType, ops, op_definer, M, V,
                                op_language, initial_types, phi_types)

EX_M = ProjType(name="EX_M", mode="M")
EX_TY = ProjType(name="EX_TY", mode="V")
EX_TB = ProjType(name="EX_TB", mode="V")
EX_V = ProjType(name="EX_V", mode="V")
T = ProjType(name="T", mode="M")
F = ProjType(name="F", mode="M")

standard_proj = [M, EX_M, EX_TY, EX_TB, EX_V]
value_proj = [M, V, EX_M, EX_TY, EX_TB, EX_V]

op = op_definer("entangled.python", standard_proj)

python_operations = ops([
    op('BEFORE_ASYNC_WITH', []),
    op('BUILD_LIST_UNPACK', []),
    #op('BUILD_MAP_UNPACK', []),
    #op('BUILD_MAP_UNPACK_WITH_CALL', []),
    op('BUILD_SET', []),
    #op('BUILD_SET_UNPACK', []),
    op('BUILD_TUPLE_UNPACK', []),
    op('COMPARE_OP', [], value_proj),
    op('DELETE_DEREF', []),
    op('END_FINALLY', []),
    op('EXTENDED_ARG', []),
    op('GET_AITER', []),
    op('GET_AWAITABLE', []),
    op('GET_ITER', ['iterable'], value_proj),
    op('GET_YIELD_FROM_ITER', []),
    op('IMPORT_FROM', []),
    op('IMPORT_NAME', []),
    op('LOAD_BUILD_CLASS', [], value_proj), # builtins['__build_class__']
    op('LOAD_CLASSDEREF', [], value_proj),
    op('LOAD_FAST', ['local_name', 'value'], value_proj),
    op('SETUP_ASYNC_WITH', []),
    op('SETUP_EXCEPT', ['target', 'stack_size']),
    op('SETUP_FINALLY', ['target', 'stack_size']),
    op('SETUP_LOOP', []),
    op('STORE_FAST', ['local_name', 'value'], value_proj),
    op('UNARY_INVERT', ['value'], value_proj),
    op('UNARY_NEGATIVE', [], value_proj),
    op('UNARY_NOT', [], value_proj),
    op('UNARY_POSITIVE', [], value_proj),
    op('UNPACK_EX', []),
    op('WITH_CLEANUP_FINISH', []),
    op('WITH_CLEANUP_START', []),

    op('DELETE_SUBSCR', ['base', 'index', 'value']),
    op('STORE_SUBSCR', ['base', 'index', 'value']),
    op('PRINT_EXPR', ['expr']),
    op('PRINT_ITEM_TO', ['item', 'stream']), # stream is above on the stack
    op('PRINT_ITEM', ['item']),
    op('PRINT_NEWLINE_TO', ['stream']),
    op('PRINT_NEWLINE', []),

    op('BINARY_ADD', ['left', 'right'], value_proj),
    op('BINARY_AND', ['left', 'right'], value_proj),
    op('BINARY_FLOOR_DIVIDE', ['left', 'right'], value_proj),
    op('BINARY_LSHIFT', ['left', 'right'], value_proj),
    op('BINARY_MATRIX_MULTIPLY', ['left', 'right'], value_proj),
    op('BINARY_MODULO', ['left', 'right'], value_proj),
    op('BINARY_MULTIPLY', ['left', 'right'], value_proj),
    op('BINARY_OR', ['left', 'right'], value_proj),
    op('BINARY_POWER', ['left', 'right'], value_proj),
    op('BINARY_RSHIFT', ['left', 'right'], value_proj),
    op('BINARY_SUBSCR', ['left', 'right'], value_proj),
    op('BINARY_SUBTRACT', ['left', 'right'], value_proj),
    op('BINARY_TRUE_DIVIDE', ['left', 'right'], value_proj),
    op('BINARY_XOR', ['left', 'right'], value_proj),
    op('BUILD_CLASS', ['name', 'bases', 'dict']),
    op('BUILD_LIST', [], value_proj, star="items"),
    op('INPLACE_ADD', ['left', 'right'], value_proj),
    op('INPLACE_AND', ['left', 'right'], value_proj),
    op('INPLACE_FLOOR_DIVIDE', ['left', 'right'], value_proj),
    op('INPLACE_LSHIFT', ['left', 'right'], value_proj),
    op('INPLACE_MATRIX_MULTIPLY', ['left', 'right'], value_proj),
    op('INPLACE_MODULO', ['left', 'right'], value_proj),
    op('INPLACE_MULTIPLY', ['left', 'right'], value_proj),
    op('INPLACE_OR', ['left', 'right'], value_proj),
    op('INPLACE_POWER', ['left', 'right'], value_proj),
    op('INPLACE_RSHIFT', ['left', 'right'], value_proj),
    op('INPLACE_SUBTRACT', ['left', 'right'], value_proj),
    op('INPLACE_TRUE_DIVIDE', ['left', 'right'], value_proj),
    op('INPLACE_XOR', ['left', 'right'], value_proj),


    op('UNPACK_SEQUENCE', ['xs']),
    op('STORE_ATTR', ['name', 'base', 'value']),
    op('DELETE_ATTR', ['name', 'base']),
    op('LOAD_ATTR', ['name', 'base'], value_proj),
    op('STORE_GLOBAL', ['name', 'value']),
    op('DELETE_GLOBAL', ['name']),
    op('LOAD_GLOBAL', ['name'], value_proj),


    op('BREAK_LOOP', []),
    op('BUILD_MAP', [], value_proj),
    op('BUILD_MAP.A', [], value_proj, star="items"), # any number of items
    op('BUILD_SLICE', ['start', 'stop', 'step'], value_proj),
    op('BUILD_TUPLE', [], value_proj, star="items"),
    op('CALL_FUNCTION', ['target', 'self', 'pos_args', 'kw_list', 'varargs',
                         'varkw'], value_proj),
    op('CONTINUE_LOOP', ['oparg'], [M]),
    op('DELETE_FAST', ['local_name']),
    op('DELETE_NAME', ['name']),
    op('DELETE_SLICE', ['kind', 'base', 'start', 'stop', 'step']),
    op('EXEC_STMT', ['u', 'v', 'w', 'locals']),
    op('FOR_ITER', ['loop_end', 'iterator'], value_proj),
    op('GET_ANEXT', ['awaitable'], value_proj),
    op('IMPORT_STAR', ['mod']),
    op('JUMP', ['condition'], [T, F] + standard_proj),
    op('JUMP_IF_NOT_DEBUG', [], [T, F] + standard_proj),
    op('LIST_APPEND', ['xs', 'x'], [M]),
    op('LOAD_CLOSURE', ['freevari', 'var'], value_proj),
    op('LOAD_CONST', ['oparg', 'value'], value_proj),
    op('LOAD_DEREF', ['freevar', 'value'], value_proj),
    op('LOAD_LOCALS', [], value_proj),
    op('LOAD_NAME', ['namei'], value_proj),
    op('LOOKUP_METHOD', ['namei', 'base'], value_proj),
    op('MAKE_CLOSURE', ['flags', 'code', 'closure', 'defaults']),
    op('MAKE_FUNCTION', ['flags', 'code', 'defaults'], value_proj),
    op('MAP_ADD', ['map', 'key', 'value'], standard_proj),
    op('NONE', [], value_proj),
    op('RAISE_VARARGS', [], [EX_M, EX_TY, EX_TB, EX_V]),
    op('RETURN_VALUE', [], [V, M]),
    op('SETUP_WITH', ['oparg', 'stack_depth', 'context_manager'],
       [ProjType('enter', 'V'), ProjType('exit', 'V')] + standard_proj),
    op('SET_ADD', ['xs', 'x']),
    op('SLICE', ['xs', 'start', 'stop'], value_proj),
    op('STORE_MAP', ['map', 'key', 'value']),
    op('STORE_NAME', ['namei', 'value']),
    op('STORE_SLICE', ['xs', 'start', 'stop', 'x']),
    op('STORE_DEREF', ['freevar', 'value']),
    op('YIELD_FROM', ['iterable']),
    op('YIELD_VALUE', ['value'], value_proj),

#    op('WHY_NOT', [], []),
    op('WHY_EXCEPTION', ['key', 'ex_ty', 'ex_tb', 'ex_v'], []),
#    op('WHY_RERAISE', ['key', 'ex_ty', 'ex_tb', 'ex_v'], []),
    op('WHY_RETURN', ['key'], []),
    op('WHY_BREAK', [], []),
    op('WHY_CONTINUE', [], []),
    op('WHY_YIELD', ['key', 'retval'], []),
])

optypes = op_language([initial_types, phi_types, python_operations])
