# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.op_types import op_definer, op_language, C, V, M, T, F, ProjType


P_RAISE = ProjType('raise', 'C')
P_BREAK = ProjType('break', 'C')
P_RETURN = ProjType('return', 'C')
P_CONTINUE = ProjType('return', 'C')

EX_TB = ProjType('EX_TB', 'V')
EX_TY = ProjType('EX_TY', 'V')
EX_V = ProjType('EX_V', 'V')


operations = [
    op('STORE_SUBSCR', ['base', 'index', 'value'], [C, M]),
    op('DELETE_SUBSCR', ['base', 'index'], [C, M]), # ??
    op('PRINT_EXPR', ['expr'], [C, M]),
    op('PRINT_ITEM_TO', ['stream', 'expr'], [C, M]),
    op('PRINT_ITEM', ['expr'], [C, M]),
    op('PRINT_NEWLINE_TO',['stream'], [C, M]),
    op('PRINT_NEWLINE', [], [C, M]),
    op('BUILD_CLASS', ['name', 'bases', 'attrs'], [C, V, M]),
    op('UNPACK_SEQUENCE', ['seq'], [C, V, M]),
    op('STORE_ATTR',['base', 'name', 'value'], [C, M]),
    op('DELETE_ATTR', ['name', 'base'], [C, M]),
    op('LOAD_ATTR', ['name', 'base'], [C, V, M]),
    op('STORE_GLOBAL', ['name', 'value'], [C, M]),
    op('DELETE_GLOBAL',['name'], [C, M]),
    op('LOAD_GLOBAL', ['name'], [C, V, M]),
    op('BUILD_TUPLE', [], [C, V, M], starargs="elements"),
    op('BUILD_LIST', [], [C, V, M], starargs="elements"),
    op('COMPARE_OP', ['cmp', 'x', 'y'], ['V']),
    op('IMPORT_NAME', ['name', 'fromlist', 'level'], [C, V, M]),
    op('IMPORT_FROM', ['name', 'mod'], [C, V, M]),
    op('GET_ITER', ['iterable'], [C, V, M]),
    op('GET_AITER', ['iterable'], [C, V, M]),
    op('GET_AWAITABLE', ['task'], [C, V, M]),

    op('LOAD_CONST', ['index'], [C, V, M]),
    op('LOAD_FAST', ['name', 'value'], [C, V, M]),
    op('STORE_FAST', ['name', 'value'], [C, M]),
    op('DELETE_FAST', ['name'], [C, M]),
    op('SLICE', ['base', 'start', 'stop'], [C, V, M]),
    op('STORE_SLICE', ['base', 'start', 'stop', 'value'], [C, M]),
    op('DELETE_SLICE', ['base', 'start', 'stop'], [C, M]),

    op('LIST_APPEND', ['xs', 'x'], [C, M]),
    op('SET_ADD', ['xs', 'x'], [C, M]),
    op('RAISE_VARARGS', ['ex_ty', 'ex_v', 'ex_tb'], [C, M]),
    op('LOAD_LOCALS', ['locals'], [C, V, M]),

    op('RETURN_VALUE', ['value'], [C, V, M]),
    op('YIELD_FROM', ['iterable'], [C, V, M]),
    op('YIELD_VALUE', ['value'], [C, V, M]),

    op('STORE_NAME', ['name', 'value'], [C, M]),
    op('DELETE_NAME', ['name'], [C, M]),
    op('LOAD_NAME', ['name'], [C, V, M]),
    op('LOAD_CLOSURE', ['index', 'value'], [C, V, M]),
    op('LOAD_DEREF', ['index'], [C, V, M]),
    op('STORE_DEREF', ['var', 'value'], [C, V, M]),

    op('BUILD_MAP', ['items'], [C, V, M]),
    op('STORE_MAP', ['map', 'key', 'value'], [C, M]),
    op('MAP_ADD', ['map', 'key', 'value'], [C, M]),
    op('IMPORT_STAR', ['mod'], [C, V, M]),

    op('import_star_maybe', ['import_op', 'name', 'prev'], [V]),

    # we should remove these opcodes - there's no point in building them.
    op('SETUP_LOOP', ['target', 'stack_size'], [C, V, M]),
    op('SETUP_EXCEPT', ['target', 'stack_size'], [C, V, M]),
    op('SETUP_FINALLY', ['target', 'stack_size'], [C, V, M]),
    op('SETUP_WITH', ['target', 'stack_size', 'context_manager'], [C, V, M]),
    #op('WITH_CLEANUP', ['target', 'stack_size'], [C, V, M]),
    op('POP_EXCEPT', ['target', 'stack_size'], [C, V, M]),
    op('POP_BLOCK', ['target', 'stack_size'], [C, V, M]),
    op('BEGIN_FINALLY', ['target', 'stack_size'], [C, V, M]),
    op('POP_FINALLY', ['unroller'],
       [C, V, M, P_RAISE, P_BREAK, P_RETURN, P_CONTINUE]),
    op('END_FINALLY', ['unroller'],
       [C, V, M, P_RAISE, P_BREAK, P_RETURN, P_CONTINUE]),
    op('CALL_FUNCTION',
       ['target', 'self', 'args', 'kwargs', 'args_x','kwargs_x'],
       [C, V, M]),
    op('MAKE_FUNCTION', ['flags', 'code', 'defaults'], [C, V, M]),
    op('MAKE_CLOSURE', ['flags', 'code', 'closure', 'defaults'], [C, V, M]),
    op('BUILD_SLICE', ['start', 'stop', 'slice'], [C, V, M]),
    op('JUMP_IF_NOT_DEBUG', [], [T, F, M]),
    op('JUMP', [], [T, F, M]),

    op('GET_ANEXT', ['iterator'], [C, V, M]),
    op('FOR_ITER', ['target', 'iterator'], [C, V, M]),

    op('BREAK_LOOP', [], [C, V, M]),
    op('CONTINUE_LOOP', ['target'], [C, V, M]),
    op('nop', [], [C, M]),
    op('sapplication_destructure', ['unroller'], [C, V, M]),
]

operations2 = [
]
